(* UFTP receivers *)
(* solyga@gmx.de, 2022-02-25 *)
IMPLEMENTATION MODULE Rcvrs;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Lib2, Lib3;
IMPORT UFTP, MD5, Cfg, SO, Log, Config, Files, Tmr;

CONST
  mid = Cfg.midRcvrs;

  cntMax = Cfg.cnt;

TYPE
  TCnt = [ 0 .. cntMax ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. MAX( TIdx ) ];

TYPE
  TWndLen = [ 0 .. wndMax ];
  TWndIdx = [ 0 .. wndMax - 1 ];

  TWndDesc = RECORD
    i: UFTP.TFgmIdx;  (* lower window boundary *)
    j: UFTP.TFgmIdx;  (* upper window boundary + 1 *)
    w: TWndLen;       (* window size *)
    (* rx infos *)
    d: SET OF TWndIdx;  (* done fragments, exact *)
    n: TWndLen;         (* done fragments, count *)
    (* tx infos *)
    t: ARRAY TWndIdx OF CARDINAL;  (* re-request down counter *)
  END;
  (* 0 <= i <= j <= rcvr.n *)
  (* j - i <= w <= wndMax *)

TYPE
  TRcvr = POINTER TO TRcvrDesc;
  TRcvrDesc = RECORD
    i: TIdx;           (* 0 <==> free *)
    l: UFTP.TFgmLen;
    f: Files.TFile;
    n: UFTP.TFgmIdx;  (* fgm cnt *)
    z: UFTP.TFgmLen;  (* last fgm *)
    w: TWndDesc;
    e: TErrCnts;
  END;

  TRcvrsDesc = RECORD
    n: TCnt;
    a: ARRAY TVldIdx OF TRcvrDesc;
  END;

VAR
  rcv: TRcvrsDesc;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
  HALT;  (* don't remove this before having checked return code implications *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE IncErr( r: TRcvr; e: TErr );
BEGIN
  Lib2.IncL( r^.e[ e ], 1 );
  Lib2.IncL( r^.e[ eNull ], 1 );
END IncErr;


PROCEDURE Alloc( l: UFTP.TFgmLen; f: Files.TFile; w: CARDINAL ): TRcvr;
VAR
  i: TVldIdx;
  r: TRcvr;
  t: UFTP.TPosDesc;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Alloc rcv ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' fgl ' );
    SO.WrCard( b, l, 1 );
    SO.WrStr( b, ' cnt ' );
    SO.WrLngCard( b, r^.n, 1 );
    SO.WrStr( b, ' wnd ' );
    SO.WrCard( b, w, 1 );
    Message( s );
  END Show;

BEGIN
  IF ( rcv.n >= cntMax ) OR ( l = 0 ) OR ( l > UFTP.lenMaxFgm ) OR ( f = Files.fNull ) THEN RETURN NIL END;
  IF w = 0 THEN w := 1 END;
  IF w > wndMax THEN w := wndMax END;
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    r := ADR( rcv.a[i] );
    IF r^.i = 0 THEN
      Lib2.Fill( r^, 0 );
      r^.i := i;
      r^.l := l;
      r^.f := f;
      Files.GetTgt( f, t );
      r^.n := t.l DIV l + ORD( t.l MOD l > 0 );
      r^.z := t.l MOD l;
      r^.w.w := w;
      INC( rcv.n );
      Show();
      RETURN r;
    END;
  END;
  Fatal( 'Alloc cnt inconsistency' );
END Alloc;

PROCEDURE GetFgl( r: TRcvr; VAR l: UFTP.TFgmLen );
BEGIN
  l := r^.l;
END GetFgl;

PROCEDURE Receive( r: TRcvr; i: UFTP.TFgmIdx; x-: ARRAY OF BYTE; l: CARDINAL );
VAR
  m: UFTP.TFgmIdx;
  k: TWndIdx;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Receive rcv ' );
    SO.WrCard( b, r^.i, 1 );
    SO.WrStr( b, ' idx ' );
    SO.WrLngCard( b, i, 1 );
    SO.WrStr( b, ' len ' );
    SO.WrCard( b, l, 1 );
    Message( s );
  END Show;

BEGIN
  IF ( i < r^.w.i ) OR ( i >= r^.w.j ) THEN
    IncErr( r, eWnd );
    RETURN;
  END;
  m := r^.n - 1;  (* last fgm *)
  IF ( i < m ) & ( l # r^.l ) OR
     ( i = m ) & ( l # r^.z ) THEN
    IncErr( r, eLen );
    RETURN;
  END;
  k := i MOD wndMax;
  IF k IN r^.w.d THEN
    IncErr( r, eDup );
    RETURN;
  END;
  Files.Write( r^.f, i * r^.l, x, l );
  INCL( r^.w.d, k );
  INC( r^.w.n );
  Show();
END Receive;

PROCEDURE Update( r: TRcvr; n, to: CARDINAL; p: TSndProc; f: TRef ): BOOLEAN;  (* busy *)
VAR
  i, j: UFTP.TFgmIdx;
  k: TWndIdx;
  s: UFTP.TFgmSet;
  b: UFTP.TFgmIdx;

  PROCEDURE Request( i: UFTP.TFgmIdx );
  BEGIN
    IF s = UFTP.TFgmSet{} THEN b := i END;
    IF i - b > MAX( UFTP.TRelFgmIdx ) THEN
      p( f, b, s );
      s := UFTP.TFgmSet{};
      b := i;
    END;
    INCL( s, i - b );
  END Request;

  PROCEDURE ShowTO( i: UFTP.TFgmIdx );
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrStr( b, 'Update TO rcv ' );
    SO.WrCard( b, r^.i, 1 );
    SO.WrStr( b, ' idx ' );
    SO.WrLngCard( b, i, 1 );
    SO.WrStr( b, ' to ' );
    SO.WrLngCard( b, to, 1 );
    Message( s );
  END ShowTO;

BEGIN
  (* clear left = progress w.i *)
  i := r^.w.i;
  k := i MOD wndMax;
  WHILE ( i < r^.w.j ) & ( k IN r^.w.d ) DO
    EXCL( r^.w.d, k );
    DEC( r^.w.n );
    INC( i );
    k := i MOD wndMax;
  END;
  r^.w.i := i;
  s := UFTP.TFgmSet{};
  b := 0;
  (* re-request *)
  WHILE ( i < r^.w.j ) & ( n > 0 ) DO
    k := i MOD wndMax;
    IF ~ ( k IN r^.w.d ) THEN
      IF r^.w.t[ k ] > 0 THEN DEC( r^.w.t[ k ] ) END;
      IF r^.w.t[ k ] = 0 THEN
        IncErr( r, eTO );
        ShowTO( i );
        Request( i );
        r^.w.t[ k ] := to;
        DEC( n );
      END;
    END;
    INC( i );
  END;
  (* adjust w.w (maybe later) *)
  (* request new = progress w.j *)
  j := r^.w.i + r^.w.w;
  IF j > r^.n THEN j := r^.n END;
  i := r^.w.j;
  WHILE ( i < j ) & ( n > 0 ) DO
    k := i MOD wndMax;
    Request( i );
    r^.w.t[ k ] := to;
    DEC( n );
    INC( i );
  END;
  r^.w.j := i;
  IF s # UFTP.TFgmSet{} THEN p( f, b, s ) END;
  RETURN r^.w.i < r^.n;
END Update;

PROCEDURE Free( VAR r: TRcvr );

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Free rcv ' );
    SO.WrCard( b, r^.i, 1 );
    Message( s );
  END Show;

BEGIN
  Show();
  Lib2.Fill( r^, 0 );
  r := NIL;
END Free;


PROCEDURE Show();
CONST
  ena = TRUE;
VAR
  b: SO.TBuffer;
  s: ARRAY [1..128] OF CHAR;
BEGIN
  IF ~ ena THEN RETURN END;
  SO.Init( b, s );
  SO.WrStr( b, 'InitModule' );
  SO.WrStr( b, ' cntMax ' );
  SO.WrCard( b, cntMax, 1 );
  SO.WrStr( b, ', SIZE( TRcvrDesc ) = ' );
  SO.WrCard( b, SIZE( TRcvrDesc ), 1 );
  Message( s );
END Show;

BEGIN
  Log.Register( mid, 'Rcvrs' );
  Lib2.Fill( rcv, 0 );
  Show();
END Rcvrs.
